@@ -13,10 +13,13 @@ DWORD WinNT()
     dwPlatformId=WinVer.dwPlatformId;
     dwMajorVersion=WinVer.dwMajorVersion;
     dwMinorVersion=WinVer.dwMinorVersion;
+
   }
   DWORD Result=0;
   if (dwPlatformId==VER_PLATFORM_WIN32_NT)
     Result=dwMajorVersion*0x100+dwMinorVersion;
-  return(Result);
+
+
+  return Result;
 }
 #endif
