@@ -25,29 +25,36 @@ bool MergeArchive(Archive &Arc,ComprDataIO *DataIO,bool ShowFileName,wchar Comma
       uiMsg(UIERROR_CHECKSUMPACKED, Arc.FileName, hd->FileName);
   }
 
+  bool PrevVolEncrypted=Arc.Encrypted;
+
   int64 PosBeforeClose=Arc.Tell();
 
   if (DataIO!=NULL)
-    DataIO->ProcessedArcSize+=Arc.FileLength();
+    DataIO->ProcessedArcSize+=DataIO->LastArcSize;
 
 
   Arc.Close();
 
   wchar NextName[NM];
-  wcscpy(NextName,Arc.FileName);
+  wcsncpyz(NextName,Arc.FileName,ASIZE(NextName));
   NextVolumeName(NextName,ASIZE(NextName),!Arc.NewNumbering);
 
 #if !defined(SFX_MODULE) && !defined(RARDLL)
   bool RecoveryDone=false;
 #endif
-  bool FailedOpen=false,OldSchemeTested=false;
+  bool OldSchemeTested=false;
 
-#if !defined(GUI) && !defined(SILENT)
-  // In -vp mode we force the pause before next volume even if it is present
-  // and even if we are on the hard disk. It is important when user does not
-  // want to process partially downloaded volumes preliminary.
-  if (Cmd->VolumePause && !uiAskNextVolume(NextName,ASIZE(NextName)))
-    FailedOpen=true;
+  bool FailedOpen=false; // No more next volume open attempts if true.
+#if !defined(SILENT)
+  if (Cmd->VolumePause)
+  {
+    // If next volume can't be opened exclusively, it might be still
+    // downloading, so in -vp mode user may prefer to pause until completion
+    // even if volume is exist. FMF_OPENEXCLUSIVE works in Windows only.
+    File TestOpen;
+    if (!TestOpen.Open(NextName,FMF_OPENEXCLUSIVE) && !uiAskNextVolume(NextName,ASIZE(NextName)))
+      FailedOpen=true;
+  }
 #endif
 
   uint OpenMode = Cmd->OpenShared ? FMF_OPENSHARED : 0;
@@ -67,12 +74,12 @@ bool MergeArchive(Archive &Arc,ComprDataIO *DataIO,bool ShowFileName,wchar Comma
         // Checking for new style volumes renamed by user to old style
         // name format. Some users did it for unknown reason.
         wchar AltNextName[NM];
-        wcscpy(AltNextName,Arc.FileName);
+        wcsncpyz(AltNextName,Arc.FileName,ASIZE(AltNextName));
         NextVolumeName(AltNextName,ASIZE(AltNextName),true);
         OldSchemeTested=true;
         if (Arc.Open(AltNextName,OpenMode))
         {
-          wcscpy(NextName,AltNextName);
+          wcsncpyz(NextName,AltNextName,ASIZE(NextName));
           break;
         }
       }
@@ -93,13 +100,11 @@ bool MergeArchive(Archive &Arc,ComprDataIO *DataIO,bool ShowFileName,wchar Comma
       }
 #endif
 
-#ifndef GUI
       if (!Cmd->VolumePause && !IsRemovable(NextName))
       {
         FailedOpen=true;
         break;
       }
-#endif
 #ifndef SILENT
       if (Cmd->AllYes || !uiAskNextVolume(NextName,ASIZE(NextName)))
 #endif
@@ -129,6 +134,16 @@ bool MergeArchive(Archive &Arc,ComprDataIO *DataIO,bool ShowFileName,wchar Comma
     return false;
 #endif
 
+  if (Arc.Encrypted!=PrevVolEncrypted)
+  {
+    // There is no legitimate reason for encrypted header state to be
+    // changed in the middle of volume sequence. So we abort here to prevent
+    // replacing an encrypted header volume to unencrypted and adding
+    // unexpected files by third party to encrypted extraction.
+    uiMsg(UIERROR_BADARCHIVE,Arc.FileName);
+    ErrHandler.Exit(RARX_FATAL);
+  }
+
   if (SplitHeader)
     Arc.SearchBlock(HeaderType);
   else
@@ -138,14 +153,12 @@ bool MergeArchive(Archive &Arc,ComprDataIO *DataIO,bool ShowFileName,wchar Comma
     Arc.ConvertAttributes();
     Arc.Seek(Arc.NextBlockPos-Arc.FileHead.PackSize,SEEK_SET);
   }
-#ifndef GUI
-  if (ShowFileName)
+  if (ShowFileName && !Cmd->DisableNames)
   {
     mprintf(St(MExtrPoints),Arc.FileHead.FileName);
     if (!Cmd->DisablePercentage)
       mprintf(L"     ");
   }
-#endif
   if (DataIO!=NULL)
   {
     if (HeaderType==HEAD_ENDARC)
@@ -155,10 +168,9 @@ bool MergeArchive(Archive &Arc,ComprDataIO *DataIO,bool ShowFileName,wchar Comma
       DataIO->UnpVolume=hd->SplitAfter;
       DataIO->SetPackedSizeToRead(hd->PackSize);
     }
-#ifdef SFX_MODULE
-    DataIO->UnpArcSize=Arc.FileLength();
-#endif
-    
+
+    DataIO->AdjustTotalArcSize(&Arc);
+      
     // Reset the size of packed data read from current volume. It is used
     // to display the total progress and preceding volumes are already
     // compensated with ProcessedArcSize, so we need to reset this variable.
@@ -175,13 +187,6 @@ bool MergeArchive(Archive &Arc,ComprDataIO *DataIO,bool ShowFileName,wchar Comma
 
 
 #ifdef RARDLL
-#if defined(RARDLL) && defined(_MSC_VER) && !defined(_WIN_64)
-// Disable the run time stack check for unrar.dll, so we can manipulate
-// with ChangeVolProc call type below. Run time check would intercept
-// a wrong ESP before we restore it.
-#pragma runtime_checks( "s", off )
-#endif
-
 bool DllVolChange(RAROptions *Cmd,wchar *NextName,size_t NameSize)
 {
   bool DllVolChanged=false,DllVolAborted=false;
@@ -189,7 +194,7 @@ bool DllVolChange(RAROptions *Cmd,wchar *NextName,size_t NameSize)
   if (Cmd->Callback!=NULL)
   {
     wchar OrgNextName[NM];
-    wcscpy(OrgNextName,NextName);
+    wcsncpyz(OrgNextName,NextName,ASIZE(OrgNextName));
     if (Cmd->Callback(UCM_CHANGEVOLUMEW,Cmd->UserData,(LPARAM)NextName,RAR_VOL_ASK)==-1)
       DllVolAborted=true;
     else
@@ -199,7 +204,7 @@ bool DllVolChange(RAROptions *Cmd,wchar *NextName,size_t NameSize)
       {
         char NextNameA[NM],OrgNextNameA[NM];
         WideToChar(NextName,NextNameA,ASIZE(NextNameA));
-        strcpy(OrgNextNameA,NextNameA);
+        strncpyz(OrgNextNameA,NextNameA,ASIZE(OrgNextNameA));
         if (Cmd->Callback(UCM_CHANGEVOLUME,Cmd->UserData,(LPARAM)NextNameA,RAR_VOL_ASK)==-1)
           DllVolAborted=true;
         else
@@ -216,28 +221,7 @@ bool DllVolChange(RAROptions *Cmd,wchar *NextName,size_t NameSize)
   {
     char NextNameA[NM];
     WideToChar(NextName,NextNameA,ASIZE(NextNameA));
-    // Here we preserve ESP value. It is necessary for those developers,
-    // who still define ChangeVolProc callback as "C" type function,
-    // even though in year 2001 we announced in unrar.dll whatsnew.txt
-    // that it will be PASCAL type (for compatibility with Visual Basic).
-#if defined(_MSC_VER)
-#ifndef _WIN_64
-    __asm mov ebx,esp
-#endif
-#elif defined(_WIN_ALL) && defined(__BORLANDC__)
-    _EBX=_ESP;
-#endif
     int RetCode=Cmd->ChangeVolProc(NextNameA,RAR_VOL_ASK);
-
-    // Restore ESP after ChangeVolProc with wrongly defined calling
-    // convention broken it.
-#if defined(_MSC_VER)
-#ifndef _WIN_64
-    __asm mov esp,ebx
-#endif
-#elif defined(_WIN_ALL) && defined(__BORLANDC__)
-    _ESP=_EBX;
-#endif
     if (RetCode==0)
       DllVolAborted=true;
     else
@@ -272,21 +256,10 @@ bool DllVolNotify(RAROptions *Cmd,wchar *NextName)
   }
   if (Cmd->ChangeVolProc!=NULL)
   {
-#if defined(_WIN_ALL) && !defined(_MSC_VER) && !defined(__MINGW32__)
-    _EBX=_ESP;
-#endif
     int RetCode=Cmd->ChangeVolProc(NextNameA,RAR_VOL_NOTIFY);
-#if defined(_WIN_ALL) && !defined(_MSC_VER) && !defined(__MINGW32__)
-    _ESP=_EBX;
-#endif
     if (RetCode==0)
       return false;
   }
   return true;
 }
-
-#if defined(RARDLL) && defined(_MSC_VER) && !defined(_WIN_64)
-// Restore the run time stack check for unrar.dll.
-#pragma runtime_checks( "s", restore )
-#endif
 #endif
